"""GLUDB provides a fairly simple way to read/write data to some popular
datastores like Amazon's DynamoDB and Google Cloud Datastore. We provide:

* A simple abstraction layer for annotating classes that should be stored in
  the database
* Support for versioning by automatically storing change history with the data
* Automated "indexing", which includes querying on the value of indexes
* Automated, configurable backup to Amazon's S3 (and Glacier depending on how
  you configure the S3 buckets)

We currently support Python 2 (2.7 and greater) and 3 (3.4 and greater). The
data stores currently supported are:

* sqlite
* DynamoDB
* Google Cloud Datastore
* MongoDB
"""



"""gludb.simple

Provides the simplest possible interface to our functionality.

We provide a simple annotation to create classes with fields (with optional
default values), parameterized constructors, persistence, and data operations.
You are free to derive from any object you wish. See gludb.data if you need
custom or more advanced functionality.

    @DBObject
    class Demo(object):
        some_field = Field()
        my_number = Field(default=42)

    d = Demo(some_field='foo', my_number=3.14)
    print(d.to_data())  # Prints a JSON representation
    d1 = Demo.from_data(d.to_data())  # Clone using persistence functions
    d.save()  # Save to database
    for obj in Demo.find_all():  # Print json rep of all objects in DB
        print(obj.to_data())

Also note that currently we aren't supporting nested DBObject objects.
HOWEVER, we make no restrictions on a field being a JSON-compatible Python
type. We make it possible to supply a decent default value by allowing a
function to be specified as a default value - it will be called when a default
value is needed. For example:

    @DBObject
    class Complicated(object):
        name = Field(default='')
        complex_data = Field(default=dict)

    c = Complicate(name)
    c.complex_data['a'] = 123
    c.complex_data['b'] = 456

IMPORTANT: you should *NOT* just use a default object like this:
`Field(default={})`. Modifications made to the default object will become the
NEW default for other classes. See
[here](http://effbot.org/zone/default-values.htm). However, you may supply
a function that will be called to retreive a default value. In this example
you should use `Field(default=dict)`.
"""

import json

from .utils import now_field
from .data import Storable, DatabaseEnabled, orig_version
from .versioning import VersioningTypes, record_diff, append_diff_hist


class _NO_VAL:
    """Simple helper we use instead of None (in case they want to use a default
    value of None)."""
    pass


class Field(object):
    """Support for class-level field declaration.
    """
    def __init__(self, default=''):
        self.name = None
        self.default = default

    def get_default_val(self):
        """Helper to expand default value (support callables)"""
        val = self.default
        while callable(val):
            val = val()
        return val


def _auto_init(self, *args, **kwrds):
    """Our decorator will add this as __init__ to target classes
    """
    for fld in getattr(self, '__fields__', []):
        val = kwrds.get(fld.name, _NO_VAL)
        if val is _NO_VAL:
            val = fld.get_default_val()
        setattr(self, fld.name, val)

    if callable(getattr(self, 'setup', None)):
        self.setup(*args, **kwrds)

# Sneaky trick - we tag __init__ functions that are Ok to be overwritten...
# this is mainly for Python 2 (in Python 3 we can successfully do an equality
# comparison between _auto_init and a replaced __init__)
_auto_init._clobber_ok = True

# Actual check for an overridable __init__ given a class
def ctor_overridable(cls):
    prev_init = getattr(cls, "__init__", None)
    if not callable(prev_init):
        return True
    if prev_init in [object.__init__, _auto_init]:
        return True
    if getattr(prev_init, '_clobber_ok', False):
        return True

    print(cls, prev_init, getattr(prev_init, '_clobber_ok', 'missing'))
    return False  # Not on our list


def _get_table_name(cls):
    return cls.__table_name__


def _get_id(self):
    return self.id


def _set_id(self, new_id):
    self.id = new_id


def _to_data(self):
    def getval(fld):
        val = getattr(self, fld.name, _NO_VAL)
        if val is _NO_VAL:
            val = fld.get_default_val()
        return val

    # Update the datetime fields that we add automatically
    now = now_field()
    setattr(self, '_last_update', now)
    if not getattr(self, '_create_date', ''):
        setattr(self, '_create_date', now)

    data = dict([(fld.name, getval(fld)) for fld in self.__fields__])

    return json.dumps(data)


def _from_data(cls, data):
    data_dict = json.loads(data)
    return cls(**data_dict)


def _index_names(cls):
    def is_index(name):
        attr = getattr(cls, name, None)
        return getattr(attr, 'is_index', False)

    return [name for name in dir(cls) if is_index(name)]


def _indexes(self):
    def get_val(name):
        attr = getattr(self, name, None)
        while callable(attr):
            attr = attr()
        return attr

    return dict([
        (name, get_val(name)) for name in self.__class__.index_names()
    ])


def _delta_save(save_method):
    def wrapper(self):
        # Get the diff's being saved
        pre_changes = orig_version(self)
        curr_data = self.to_data()
        diff = record_diff(pre_changes, curr_data) if pre_changes else None

        # Need to save changes?
        if diff:
            ver_hist = self.get_version_hist()
            ver_hist = append_diff_hist(diff, ver_hist)
            setattr(self, '_version_hist', ver_hist)

        return save_method(self)

    return wrapper


def _get_version_hist(self):
    if self.__versioning__ == VersioningTypes.NONE:
        return None
    elif self.__versioning__ == VersioningTypes.DELTA_HISTORY:
        return getattr(self, '_version_hist', list())
    else:
        raise ValueError("Unknown versioning type")


def DBObject(table_name, versioning=VersioningTypes.NONE):
    """Classes annotated with DBObject gain persistence methods."""
    def wrapped(cls):
        field_names = set()
        all_fields = []

        for name in dir(cls):
            fld = getattr(cls, name)
            if fld and isinstance(fld, Field):
                fld.name = name
                all_fields.append(fld)
                field_names.add(name)

        def add_missing_field(name, default='', insert_pos=None):
            if name not in field_names:
                fld = Field(default=default)
                fld.name = name
                all_fields.insert(
                    len(all_fields) if insert_pos is None else insert_pos,
                    fld
                )

        add_missing_field('id', insert_pos=0)
        add_missing_field('_create_date')
        add_missing_field('_last_update')
        if versioning == VersioningTypes.DELTA_HISTORY:
            add_missing_field('_version_hist', default=list)

        # Things we count on as part of our processing
        cls.__table_name__ = table_name
        cls.__versioning__ = versioning
        cls.__fields__ = all_fields

        # Give them a ctor for free - but make sure we aren't clobbering one
        if not ctor_overridable(cls):
            raise TypeError(
                'Classes with user-supplied __init__ should not be decorated '
                'with DBObject. Use the setup method'
            )
        cls.__init__ = _auto_init

        # Duck-type the class for our data methods
        cls.get_table_name = classmethod(_get_table_name)
        cls.get_id = _get_id
        cls.set_id = _set_id
        cls.to_data = _to_data
        cls.from_data = classmethod(_from_data)
        cls.index_names = classmethod(_index_names)
        cls.indexes = _indexes
        # Bonus methods they get for using gludb.simple
        cls.get_version_hist = _get_version_hist

        # Register with our abc since we actually implement all necessary
        # functionality
        Storable.register(cls)

        # And now that we're registered, we can also get the database
        # read/write functionality for free
        cls = DatabaseEnabled(cls)

        if versioning == VersioningTypes.DELTA_HISTORY:
            cls.save = _delta_save(cls.save)

        return cls

    return wrapped


def Index(func):
    """Marks instance methods of a DBObject-decorated class as being used for
    indexing. The function name is used as the index name, and the return
    value is used as the index value.

    Note that callables are call recursively so in theory you can return
    a function which will be called to get the index value"""
    func.is_index = True
    return func



"""gludb.backends.dynamodb - backend dynamodb database module
"""

import os

import boto.exception
import boto.dynamodb2  # NOQA

from boto.dynamodb2.layer1 import DynamoDBConnection
from boto.dynamodb2.table import Table
from boto.dynamodb2.fields import HashKey, GlobalIncludeIndex
from boto.dynamodb2.items import Item
from boto.dynamodb2.exceptions import ResourceNotFoundException, ItemNotFound
from boto.exception import JSONResponseError

from ..utils import uuid


def get_conn():
    """Return a connection to DynamoDB (and handle local/debug
    possibilities)"""
    if os.environ.get('DEBUG', False) or os.environ.get('travis', False):
        # In DEBUG mode - use the local DynamoDB
        # This also works for travis since we'll be running dynalite
        conn = DynamoDBConnection(
            host='localhost',
            port=8000,
            aws_access_key_id='TEST',
            aws_secret_access_key='TEST',
            is_secure=False
        )
    else:
        # Regular old production
        conn = DynamoDBConnection()

    return conn


def gsi_name(index_name):
    """Standardize how we create a GSI name for DynamoDB from the given index
    name from the class"""
    return index_name + '_gsiindex'


class DynamoMappings(object):
    """DynamoDB has some opinions about what you can store or query in an
    attribute. We're going to use mappings to fix that.
    """

    # Yes, the bracket-ish map vals are illegal in just about any language that
    # uses bracket-type syntax. That's on purpose.
    NONE_VAL = "<{[None>}]"
    EMPTY_STR_VAL = "<{[''>}]"

    @staticmethod
    def map_index_val(index_val):
        """Xform index_val so that it can be stored/queried"""
        if index_val is None:
            return DynamoMappings.NONE_VAL

        index_val = str(index_val)
        if not index_val:
            return DynamoMappings.EMPTY_STR_VAL

        return index_val

    @staticmethod
    def unmap_stored_val(stored_val):
        """Inverse of index_val_mapping. Note that we currently don't use it
        because we don't actually read back index values (since they are
        generated by Python functions)"""
        if stored_val == DynamoMappings.NONE_VAL:
            return None
        elif stored_val == DynamoMappings.EMPTY_STR_VAL:
            return ''
        else:
            return stored_val


def delete_table(table_name):
    """Mainly for testing"""
    Table(table_name, connection=get_conn(), schema=[HashKey('id')]).delete()


class Backend(object):
    def __init__(self, **kwrds):
        pass  # No current keywords needed/used

    def table_schema_call(self, target, cls):
        """Call the callable target with the args and keywords needed for the
        table defined by cls. This is how we centralize the Table.create and
        Table ctor calls
        """
        index_defs = []
        for name in cls.index_names() or []:
            index_defs.append(GlobalIncludeIndex(
                gsi_name(name),
                parts=[HashKey(name)],
                includes=['value']
            ))

        return target(
            cls.get_table_name(),
            connection=get_conn(),
            schema=[HashKey('id')],
            global_indexes=index_defs or None
        )

    def ensure_table(self, cls):
        exists = True
        conn = get_conn()

        try:
            descrip = conn.describe_table(cls.get_table_name())
            assert descrip is not None
        except ResourceNotFoundException:
            # Expected - this is what we get if there is no table
            exists = False
        except JSONResponseError:
            # Also assuming no table
            exists = False

        if not exists:
            table = self.table_schema_call(Table.create, cls)
            assert table is not None

    def get_class_table(self, cls):
        """Return a DynamoDB table object for the given class
        """
        return self.table_schema_call(Table, cls)

    def find_one(self, cls, id):
        try:
            db_result = self.get_class_table(cls).lookup(id)
        except ItemNotFound:
            # according to docs, this shouldn't be required, but it IS
            db_result = None

        if not db_result:
            return None

        obj = cls.from_data(db_result['value'])
        return obj

    def find_all(self, cls):
        final_results = []
        table = self.get_class_table(cls)
        for db_result in table.scan():
            obj = cls.from_data(db_result['value'])
            final_results.append(obj)

        return final_results

    def find_by_index(self, cls, index_name, value):
        query_args = {
            index_name + '__eq': DynamoMappings.map_index_val(value),
            'index': gsi_name(index_name)
        }

        final_results = []
        for db_result in self.get_class_table(cls).query_2(**query_args):
            obj = cls.from_data(db_result['value'])
            final_results.append(obj)

        return final_results

    def save(self, obj):
        if not obj.id:
            obj.id = uuid()

        stored_data = {
            'id': obj.id,
            'value': obj.to_data()
        }

        index_vals = obj.indexes() or {}
        for key in obj.__class__.index_names() or []:
            val = index_vals.get(key, '')
            stored_data[key] = DynamoMappings.map_index_val(val)

        table = self.get_class_table(obj.__class__)
        item = Item(table, data=stored_data)

        item.save(overwrite=True)



"""gludb.data

The "core" functionality. If you're unsure what to use, you should look into
gludb.simple. This module is for those needing advanced functionality or
customization
"""

from abc import ABCMeta, abstractmethod

from .config import get_mapping


# A little magic for using metaclasses with both Python 2 and 3
def _with_metaclass(meta, *bases):
    """Taken from future.utils, who took is from  from jinja2/_compat.py.
    License: BSD."""
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)

    return metaclass('temporary_class', None, {})


class Storable(_with_metaclass(ABCMeta)):
    """Our abstract base class that marks subclasses as persistable and
    storable. Note that the DBObject annotation in gludb.simple registers
    all annotated classes as 'virtual base classes' of Storage so that you
    can test them with isinstance(obj, Storable)"""

    # This is field we use for the original version of the object (saved after
    # retrieval and before any edits occur)
    ORIG_VER_FIELD_NAME = "_prev_version"

    @classmethod
    @abstractmethod
    def get_table_name(self):
        """Return the name of the table/collection/etc where objects should
        be saved/loaded"""
        pass

    @abstractmethod
    def get_id(self):
        """The instance should return the current key/ID for the instance. If
        a 'falsey' value is return, on save one will be created and set via a
        call to self.set_id"""
        pass

    @abstractmethod
    def set_id(self, new_id):
        """The instance should accept a new key/ID. See also get_id"""
        pass

    @abstractmethod
    def to_data(self):
        """The instance should return JSON representation of it's internal
        state. See also from_data"""
        pass

    @classmethod
    @abstractmethod
    def from_data(self):
        """This classmethod returns a new instance of the subclass populated
        from the JSON representation"""
        pass

    @classmethod
    def index_names(self):
        """This optional method should return an iterable of index names. These
        names should correspond to the names used in the dictionary returned by
        the instance method `indexes` (below)"""
        return None

    def indexes(self):
        """This optional method should return a dictionary of index name values
        that can be used in a query. Note that this is not considered required
        data, so a backend could ignore indexes if necessary."""
        return None


def _ensure_table(cls):
    get_mapping(cls).ensure_table(cls)


def _post_load(obj):
    # Perform all necessary post load operations we want done when reading
    # from the database. We return the changed object, but make NO EFFORT
    # to keep from mutating the original object.
    if obj:
        setattr(obj, Storable.ORIG_VER_FIELD_NAME, obj.to_data())
    return obj


def _find_one(cls, id):
    return _post_load(get_mapping(cls).find_one(cls, id))


def _find_all(cls):
    return [
        _post_load(obj)
        for obj in get_mapping(cls).find_all(cls)
    ]


def _find_by_index(cls, index_name, value):
    return [
        _post_load(obj)
        for obj in get_mapping(cls).find_by_index(cls, index_name, value)
    ]


def _save(self):
    # Actual save
    get_mapping(self.__class__).save(self)

    # Now we have a new original version
    setattr(self, Storable.ORIG_VER_FIELD_NAME, self.to_data())


def DatabaseEnabled(cls):
    """Classes annotated with DatabaseEnabled gain persistence methods. All
    this really does is add some functions that forward to the mapped database
    class"""
    if not issubclass(cls, Storable):
        raise ValueError(
            "%s is not a subclass of gludb.datab.Storage" % repr(cls)
        )

    cls.ensure_table = classmethod(_ensure_table)
    cls.find_one = classmethod(_find_one)
    cls.find_all = classmethod(_find_all)
    cls.find_by_index = classmethod(_find_by_index)
    cls.save = _save

    return cls


def orig_version(obj):
    """Return the original version of an object (defined as what was read from
    the database before any user edits). If there isn't a previous version (for
    instance, newly created objects don't have a previous version), then None
    is returned. Mainly useful for testing"""
    return getattr(obj, Storable.ORIG_VER_FIELD_NAME, None)



"""Provide gludb configuration. This consists mainly of a mapping from Storable
classes to a database configuration. It also includes a default mapping for
classes not specifically mapped to a database.

We check for a mapping for a class in MRO (Method Resolution Order). Suppose
a class Busy derives from the three classes X, Y, and Z - all of which derive
from class Base:

    class Base(object):
        pass

    class X(Base):
        pass
    class Y(Base):
        pass
    class Z(Base):
        pass

    class Busy(X, Y, Z):
        pass

Then a check for mapping would first checking for class Busy, then (in order)
classes X, Y, Z, Base, and object. If this is the first time you've seen
multiple inheritance, you'll note that the order of Busy's super classes is
important. This is how Python resolves method calls (we didn't just make this
up). In fact, we depend on the results of the standard library call
`inspect.getmro`.

If none of the classes mentioned have a mapping, then the default mapping will
be used. If there is no default mapping, then the class can't be mapped to a
database instance and an error will be thrown.

Astute readers will note that you could map the class `object` to a database
as a kind of default mapping. We generally don't recommend this, because it
would work *sort of*. Some notes:

* Recall that we support Python 2 and 3. In Python 3.4, classes declared
  without a base class get `object` as a base class automatically. In Python
  2.7 they just don't have a base class. That means that using an `object`
  DB mapping won't work as a default in every case.
* We always check for the default database mapping last, so mapping to object
  would be the last thing checked before the actual default.
"""

from inspect import getmro
from importlib import import_module


class Database(object):
    """Configuration class representing a database instance supported by one
    of our backends"""
    def __init__(self, db_driver, **kwrds):
        mod = import_module('.backends.'+db_driver, __package__)
        Backend = getattr(mod, "Backend")
        self.backend = Backend(**kwrds)

    def ensure_table(self, cls):
        return self.backend.ensure_table(cls)

    def find_one(self, cls, id):
        return self.backend.find_one(cls, id)

    def find_all(self, cls):
        return self.backend.find_all(cls)

    def find_by_index(self, cls, index_name, value):
        return self.backend.find_by_index(cls, index_name, value)

    def save(self, obj):
        self.backend.save(obj)


# Note our use of a class with a singleton instance - so configuration is
# process-wide. This makes things much simpler for some users (which is one
# of our main design drivers). One potential enhancement is to allow multiple
# database mappings - perhaps as a Flask Blueprint plugin/addon
class _DatabaseMapping(object):
    def __init__(self):
        self.clear_mappings()

    def add_mapping(self, cls, db):
        self.mapping[cls] = db

    def get_mapping(self, cls, no_mapping_ok=False):
        db = None

        for candidate_cls in getmro(cls):
            db = self.mapping.get(candidate_cls, None)
            if db is not None:
                break

        if db is None:
            db = self.default_database

        if db is None:
            if no_mapping_ok:
                return None
            raise ValueError("There is no database mapping for %s" % repr(cls))

        return db

    def clear_mappings(self):
        self.default_database = None
        self.mapping = {}

_database_mapping = _DatabaseMapping()


def default_database(db):
    """Set the default database configuration used for classes without a
    specific mapping"""
    _database_mapping.default_database = db


def class_database(cls, db):
    """Map a class (for which we assume issubclass(cls, Storable)==True) to
    a database configuration"""
    _database_mapping.add_mapping(cls, db)


def clear_database_config():
    """Reset all mappings to default state. Note that any in-memory databases
    will be lost"""
    _database_mapping.clear_mappings()


def get_mapping(cls, no_mapping_ok=False):
    """Return a database config object for the given class"""
    return _database_mapping.get_mapping(cls, no_mapping_ok)



"""Central place for misc utilities
"""

import datetime

from uuid import uuid4


def uuid():
    """Return a decent UUID as a string"""
    return uuid4().hex


def now_field():
    """Return a string we use for storing our date time values"""
    return 'UTC:' + datetime.datetime.utcnow().isoformat()


def parse_now_field(s):
    """Return a datetime instance from a string generated by now_field.
    IMPORTANT: the datetime will be in UTC"""
    if not s.startswith('UTC:'):
        return None  # Invalid string
    s = s[4:]

    # isoformat can return strings both with and without microseconds - we
    # account for both
    try:
        dt = datetime.datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        dt = datetime.datetime.strptime(s, '%Y-%m-%dT%H:%M:%S')

    return dt



"""This sub-package contains all backends available in gludb. In general,
you shouldn't need to directly use any of the classes here since they will be
created and managed for you via the mappings available in gludb.config
"""



"""gludb.backends.gcd - backend Google Cloud Datastore module
"""

import sys

from ..utils import uuid

if sys.version_info >= (3, 0):
    raise ImportError("GLUDB GCD Backend only supports Python 2.7")

import googledatastore as datastore
# TODO: one day use this instead (when it has Python3 support)
# from gcloud import datastore


class DatastoreTransaction(object):
    def __init__(self):
        self.tx = None
        self.commit_req = None

    def __enter__(self):
        req = datastore.BeginTransactionRequest()
        resp = datastore.begin_transaction(req)
        self.tx = resp.transaction
        return self

    def get_commit_req(self):
        if not self.commit_req:
            self.commit_req = datastore.CommitRequest()
            self.commit_req.transaction = self.tx
        return self.commit_req

    def get_upsert(self):
        return self.get_commit_req().mutation.upsert.add()

    def __exit__(self, type, value, traceback):
        if self.commit_req:
            datastore.commit(self.commit_req)
            self.commit_req = None

        if self.tx:
            # Simple, but we might want to log or troubleshoot one day
            self.tx = None


def make_key(table_name, objid):
    key = datastore.Key()
    path = key.path_element.add()
    path.kind = table_name
    path.name = str(objid)
    return key


def write_rec(table_name, objid, data, index_name_values):
    with DatastoreTransaction() as tx:
        entity = tx.get_upsert()

        entity.key.CopyFrom(make_key(table_name, objid))

        prop = entity.property.add()
        prop.name = 'id'
        prop.value.string_value = objid

        prop = entity.property.add()
        prop.name = 'value'
        prop.value.string_value = data

        for name, val in index_name_values:
            prop = entity.property.add()
            prop.name = name
            prop.value.string_value = str(val)


def extract_entity(found):
    obj = dict()
    for prop in found.entity.property:
        obj[prop.name] = prop.value.string_value
    return obj


def read_rec(table_name, objid):
    req = datastore.LookupRequest()
    req.key.extend([make_key(table_name, objid)])

    for found in datastore.lookup(req).found:
        yield extract_entity(found)


def read_by_indexes(table_name, index_name_values=None):
    req = datastore.RunQueryRequest()

    query = req.query
    query.kind.add().name = table_name

    if not index_name_values:
        index_name_values = []
    for name, val in index_name_values:
        queryFilter = query.filter.property_filter
        queryFilter.property.name = name
        queryFilter.operator = datastore.PropertyFilter.EQUAL
        queryFilter.value.string_value = str(val)

    loop_its = 0
    have_more = True

    while have_more:
        resp = datastore.run_query(req)

        found_something = False
        for found in resp.batch.entity_result:
            yield extract_entity(found)
            found_something = True

        if not found_something:
            # This is a guard against bugs or excessive looping - as long we
            # can keep yielding records we'll continue to execute
            loop_its += 1
            if loop_its > 5:
                raise ValueError("Exceeded the excessive query threshold")

        if resp.batch.more_results != datastore.QueryResultBatch.NOT_FINISHED:
            have_more = False
        else:
            have_more = True
            end_cursor = resp.batch.end_cursor
            query.start_cursor.CopyFrom(end_cursor)


def delete_table(table_name):
    """Mainly for testing"""

    to_delete = [
        make_key(table_name, rec['id'])
        for rec in read_by_indexes(table_name, [])
    ]

    with DatastoreTransaction() as tx:
        tx.get_commit_req().mutation.delete.extend(to_delete)


class Backend(object):
    def __init__(self, **kwrds):
        pass  # No current keywords needed/used

    def ensure_table(self, cls):
        pass  # Currently nothing needs to be done

    def find_one(self, cls, id):
        db_result = None
        for rec in read_rec(cls.get_table_name(), id):
            db_result = rec
            break  # Only read the first returned - which should be all we get
        if not db_result:
            return None

        obj = cls.from_data(db_result['value'])
        return obj

    def find_all(self, cls):
        final_results = []
        for db_result in read_by_indexes(cls.get_table_name(), []):
            obj = cls.from_data(db_result['value'])
            final_results.append(obj)

        return final_results

    def find_by_index(self, cls, index_name, value):
        table_name = cls.get_table_name()
        index_name_vals = [(index_name, value)]

        final_results = []
        for db_result in read_by_indexes(table_name, index_name_vals):
            obj = cls.from_data(db_result['value'])
            final_results.append(obj)

        return final_results

    def save(self, obj):
        if not obj.id:
            obj.id = uuid()

        index_names = obj.__class__.index_names() or []
        index_dict = obj.indexes() or {}
        index_name_values = [
            (key, index_dict.get(key, ''))
            for key in index_names
        ]

        write_rec(
            obj.__class__.get_table_name(),
            obj.id,
            obj.to_data(),
            index_name_values
        )



"""Supply backup functionality for gludb. Note that we mainly provide a way for
users to create a small script that backs up their data to S3. We are assuming
that the S3 bucket used will be configured for archival (either deletion or
archival to Glacier)
"""

import sys
import os
import pkgutil
import tarfile

from importlib import import_module
from inspect import getmembers, isclass, getmro
from tempfile import NamedTemporaryFile

from boto.s3.connection import S3Connection, OrdinaryCallingFormat
from boto.s3.key import Key

from .utils import now_field
from .config import get_mapping
from .data import Storable
from .simple import DBObject, Field

# TODO: need the back to check for previous backups and only run if enough time
#       has elapsed. Best way is probably: store json for backup after
#       success. Then we read it back to get date/time of last backup


def is_backup_class(cls):
    """Return true if given class supports back up. Currently this means a
    gludb.data.Storable-derived class that has a mapping as defined in
    gludb.config"""
    return True if (
        isclass(cls) and
        issubclass(cls, Storable) and
        get_mapping(cls, no_mapping_ok=True)
    ) else False


def backup_name(cls):
    """Return a usable name for the data stored for a class"""
    return cls.__name__ + '--' + cls.get_table_name()

# write_line is used below for Python 2&3 compatibility. We provide strip_line
# as a convenience for people reading backups (since it includes implicit
# encoding)
if sys.version_info >= (3, 0):
    def write_line(file_obj, line):
        file_obj.write(bytes(str(line) + '\n', 'UTF-8'))

    def strip_line(line):
        return str(line, 'UTF-8').strip()
else:
    def write_line(file_obj, line):
        file_obj.write(str(line))
        file_obj.write('\n')

    def strip_line(line):
        return str(line).strip()

# Maintain only one copy of the doc string for functions that have different
# versions for Python 2 and 3
write_line.__doc__ = """Write the given line to the given file-like
object with a terminating linefeed character. UTF-8 encoding is assumed"""

strip_line.__doc__ = """Return a stripped (trimmed) version of the line read
back from a file assuming that the line was originally written with the
write_line function UTF-8 encoding is assumed"""


# Turns out the library qualname doesn't handle annotated classes and we need
# Python 2 for gcd docs .... so we'll force-annotated Backup below
class Backup(object):
    """This is the main interface to gludb.backup functionality. When creating
    the instance must specify a bucket_name for backups. You may optionally
    specify a name, AWS access ID, and AWS secret key. Note that if you don't
    specify AWS credentials the environment variables AWS_ACCESS_KEY_ID and
    AWS_SECRET_ACCESS_KEY will checked. If those aren't defined, then empty
    strings will be used"""

    name = Field('backup')
    timestamp = Field(now_field)
    aws_access_key = Field('')
    aws_secret_key = Field('')
    bucketname = Field('')
    backup_log = Field(list)

    def setup(self, *args, **kwrds):
        # We don't ever store this name-to-class mapping
        self.classes = dict()

        # If they didn't manually specify AWS creds when we were created,
        # then we read them from the AWS-documented environ variables that
        # boto would use anyway
        if not self.aws_access_key:
            self.aws_access_key = os.environ.get('AWS_ACCESS_KEY_ID', '')
        if not self.aws_secret_key:
            self.aws_secret_key = os.environ.get('AWS_SECRET_ACCESS_KEY', '')

    def add_package(
        self,
        pkg_name,
        recurse=True,
        include_bases=True,
        parent_pkg=None
    ):
        """Add all classes to the backup in the specified package (including
        all modules and all sub-packages) for which is_backup_class returns
        True. Note that self.add_class is used, so base classes will added as
        well.

        Parameters:
        * pkg_name - a string representing the package name. It may be
          relative _if_ parent_pkg is supplied as well
        * recurse - (default value of True) if False, sub-packages will _not_
          be examined
        * include_bases - (default value of True) is passed directly to
          add_class for every class added
        * parent_pkg - a string representing the parent package of the relative
          package specified in pkg_name. Note that you should specify
          parent_pkg _only_ if pkg_name should be interpreted as relative

        An an example of both relative and absolute package imports, these
        are equivalent:

        ````
        backup.add_package('toppackage.subpackage')
        backup.add_package('subpackage', parent_pkg='toppackage')
        ````
        """
        if parent_pkg:
            pkg = import_module('.' + pkg_name, parent_pkg)
        else:
            pkg = import_module(pkg_name)

        for module_loader, name, ispkg in pkgutil.walk_packages(pkg.__path__):
            if not ispkg:
                # Module
                mod = import_module('.' + name, pkg_name)
                for name, member in getmembers(mod):
                    if is_backup_class(member):
                        self.add_class(member, include_bases=include_bases)
            elif recurse:
                # Package and we're supposed to recurse
                self.add_package(
                    pkg_name + '.' + name,
                    recurse=True,
                    include_bases=include_bases,
                    parent_pkg=parent_pkg
                )

    def add_class(self, cls, include_bases=True):
        """Add the specified class (which should be a class object, _not_ a
        string). By default all base classes for which is_backup_class returns
        True will also be added. `include_bases=False` may be spcified to
        suppress this behavior. The total number of classes added is returned.
        Note that if is_backup_class does not return True for the class object
        passed in, 0 will be returned. If you specify include_bases=False, then
        the maximum value that can be returned is 1."""
        if not is_backup_class(cls):
            return 0

        added = 0

        cls_name = backup_name(cls)
        if cls_name not in self.classes:
            self.classes[cls_name] = cls
            self.log("Added class for backup: %s", cls_name)
            added = 1

        if include_bases:
            for candidate_cls in getmro(cls):
                if is_backup_class(cls):
                    # Note that we don't keep recursing on base classes
                    added += self.add_class(candidate_cls, include_bases=False)

        return added

    def log(self, entry, *args):
        """Append the string supplied to the log (a list of strings). If
        additional arguments are supplied, then first string is assumed to be
        a format string and the other args are used for string interpolation.
        For instance `backup.log("%d + %d == %d", 1, 1, 2)` would result in the
        string `'1 + 1 == 2'` being logged"""
        if args:
            entry = entry % args
        self.backup_log.append(entry)

    def run_backup(self):
        """The actual backup is performed. The data for all added classes is
        extracted and written to a file per class where each line (terminated
        by a line feed character) is the JSON representing a single object.
        Those files are all archived in a single gzip'ed tarball which is
        stored in the AWS S3 bucket specified when the current instance of
        Backup was created"""
        self.log("Starting backup at %s", now_field())
        self.log("Backup config object created at %s", self.timestamp)

        # Make sure we're good to go
        for fld in ['aws_access_key', 'aws_secret_key', 'bucketname']:
            val = getattr(self, fld, None)
            if not val:
                self.log("Backup cannot start: %s is a required field", fld)
                raise ValueError(self.backup_log[-1])

        # Start the compressed tarball our data is stored in
        backup_file = NamedTemporaryFile(suffix=".tar.gz")
        backup_tarfile = tarfile.open(fileobj=backup_file, mode='w:gz')

        for cls_name, cls in self.classes.items():
            self.log("Backing up %s", cls_name)

            rec_count = 0

            with NamedTemporaryFile() as record_file:
                for rec in cls.find_all():
                    write_line(record_file, rec.to_data())
                    rec_count += 1

                record_file.flush()
                backup_tarfile.add(record_file.name, arcname=cls_name+'.json')

            self.log("%s => %d records backed up", cls_name, rec_count)

        # Finalize archive
        backup_tarfile.close()
        backup_file.flush()
        backup_size = os.stat(backup_file.name)[6]

        # Figure out key name for archived file
        key_name = ('Backup_' + now_field() + '.tar.gz').replace(':', '_')

        # upload archive to s3
        if os.environ.get('DEBUG', False) or os.environ.get('travis', False):
            # Local or CI - connect to our mock s3 service
            conn = S3Connection(
                '', '',
                is_secure=False, port=8888, host='localhost',
                calling_format=OrdinaryCallingFormat()
            )
        else:
            conn = S3Connection(self.aws_access_key, self.aws_secret_key)

        bucket = conn.get_bucket(self.bucketname)
        key = Key(bucket)
        key.key = key_name

        self.log(
            "Sending %s [size=%d bytes] with key name %s",
            backup_file.name,
            backup_size,
            key_name
        )

        # TODO: should probably look into a multi-part upload for larger backup
        key.set_contents_from_filename(backup_file.name)
        self.log("Sent %s", backup_file.name)

        # All done
        backup_file.close()
        self.log("Backup completed")

        # return the bucket name and key name for the completed backup
        return self.bucketname, key_name

# TODO: return to an annotation when we can use Python3 for our doc generation
Backup = DBObject(table_name='BackupHistory')(Backup)



"""versioning.py

GLUDB versioning implementation
"""

import json
import datetime

import json_delta  # External dependency

from .utils import now_field


# Yes, this could be an enum, but we're supporting Python 2.7
class VersioningTypes(object):
    NONE = "ver:none"
    DELTA_HISTORY = "ver:delta"


# Python 2&3 compatible string testing
def _isstr(s):
    try:
        _basestring = basestring
    except NameError:
        _basestring = str
    return isinstance(s, _basestring)


# If any parameter is a string, parse it as JSON
def _norm_json_params(*args):
    return tuple([
        json.loads(param) if _isstr(param) else param
        for param in args
    ])


def record_diff(old, new):
    """Return a JSON-compatible structure capable turn the `new` record back
    into the `old` record. The parameters must be structures compatible with
    json.dumps *or* strings compatible with json.loads. Note that by design,
    `old == record_patch(new, record_diff(old, new))`"""
    old, new = _norm_json_params(old, new)
    return json_delta.diff(new, old, verbose=False)


def record_patch(rec, diff):
    """Return the JSON-compatible structure that results from applying the
    changes in `diff` to the record `rec`. The parameters must be structures
    compatible with json.dumps *or* strings compatible with json.loads. Note
    that by design, `old == record_patch(new, record_diff(old, new))`"""
    rec, diff = _norm_json_params(rec, diff)
    return json_delta.patch(rec, diff, in_place=False)


def append_diff_hist(diff, diff_hist=list()):
    """Given a diff as generated by record_diff, append a diff record to the
    list of diff_hist records."""
    diff, diff_hist = _norm_json_params(diff, diff_hist)

    if not diff_hist:
        diff_hist = list()

    diff_hist.append({'diff': diff, 'diff_date': now_field()})
    return diff_hist


def parse_diff_hist(curr_obj, diff_hist):
    """Given a diff_hist as created, appended by append_diff_hist, yield the
    versions of the object start with curr_obj and working backwards in time.
    Each instance yielded is of the form (obj, date-string) where obj is the
    JSON version of the object created by applying a diff in the
    diff history and date-string is a string representing the date/time that
    the diff was taken"""
    curr_obj, diff_hist = _norm_json_params(curr_obj, diff_hist)

    yield (json.dumps(curr_obj), None)

    last_obj = curr_obj
    for one in reversed(diff_hist):
        last_obj = record_patch(last_obj, one['diff'])
        yield json.dumps(last_obj), one['diff_date']



"""gludb.backends.sqlite - backend sqlite database module
"""

import sqlite3

from ..utils import uuid


class Backend(object):
    def __init__(self, **kwrds):
        self.filename = kwrds.get('filename', '')
        if not self.filename:
            raise ValueError('sqlite backend requires a filename parameter')

        self.conn = sqlite3.connect(self.filename)

    def ensure_table(self, cls):
        cur = self.conn.cursor()

        table_name = cls.get_table_name()
        index_names = cls.index_names() or []

        cols = ['id text primary key', 'value text']
        for name in index_names:
            cols.append(name + ' text')

        cur.execute('create table if not exists %s (%s)' % (
            table_name,
            ','.join(cols)
        ))

        for name in index_names:
            cur.execute('create index if not exists %s on %s(%s)' % (
                table_name + '_' + name + '_idx',
                table_name,
                name
            ))

        self.conn.commit()
        cur.close()

    def find_one(self, cls, id):
        found = self.find_by_index(cls, 'id', id)
        return found[0] if found else None

    def find_all(self, cls):
        return self.find_by_index(cls, '1', 1)

    def find_by_index(self, cls, index_name, value):
        cur = self.conn.cursor()

        query = 'select id,value from %s where %s = ?' % (
            cls.get_table_name(),
            index_name
        )

        found = []
        for row in cur.execute(query, (value,)):
            id, data = row[0], row[1]
            obj = cls.from_data(data)
            assert id == obj.id
            found.append(obj)

        cur.close()

        return found

    def save(self, obj):
        cur = self.conn.cursor()

        tabname = obj.__class__.get_table_name()

        index_names = obj.__class__.index_names() or []

        col_names = ['id', 'value'] + index_names
        value_holders = ','.join(['?' * len(col_names)])

        if not obj.id:
            id = uuid()
            obj.id = id

        query = 'insert or replace into %s (%s) values (%s)' % (
            tabname,
            ','.join(col_names),
            ','.join(value_holders)
        )

        values = [obj.id, obj.to_data()]

        index_vals = obj.indexes() or {}
        values += [index_vals.get(name, 'NULL') for name in index_names]

        cur.execute(query, tuple(values))
        self.conn.commit()

        cur.close()



from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
# Note that we are using README.rst - it is generated from README.md in
# build.sh
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='gludb',
    version='0.1.4',
    description='A simple database wrapper',
    long_description=long_description,
    url='https://github.com/memphis-iis/GLUDB',
    author='University of Memphis Institute for Intelligent Systems',
    author_email='cnkelly@memphis.edu',
    license='Apache Version 2.0',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Archiving :: Backup',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    keywords='database versioning backup'
             'sqlite dynamodb cloud datastore mongodb',

    packages=['gludb', 'gludb.backends'],

    install_requires=[
        "json_delta>=1.1.3",
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': [],
        'test': ['coverage', 'nose', 'tornado'],
        'dynamodb': ['boto'],
        'gcd': ['googledatastore'],
        'mongodb': ['pymongo'],
        'backups': ['boto'],
    },

    package_data={},
    data_files=[],
    entry_points={},
)



"""MongoDB backend
"""

import json

from pymongo import MongoClient
from pymongo.errors import CollectionInvalid

from ..utils import uuid


def delete_collection(db_name, collection_name, host='localhost', port=27017):
    """Almost exclusively for testing"""
    client = MongoClient("mongodb://%s:%d" % (host, port))
    client[db_name].drop_collection(collection_name)


class Backend(object):
    def __init__(self, **kwrds):
        self.mongo_url = kwrds.get('mongo_url', 'mongodb://localhost:27017')
        self.mongo_client = MongoClient(self.mongo_url)

    def get_collection(self, collection_name):
        return self.mongo_client.get_default_database()[collection_name]

    def ensure_table(self, cls):
        coll_name = cls.get_table_name()
        try:
            db = self.mongo_client.get_default_database()
            db.create_collection(coll_name)
        except CollectionInvalid:
            pass  # Expected if collection already exists

        # Make sure we have indexes
        coll = self.get_collection(coll_name)
        for idx_name in cls.index_names():
            coll.ensure_index(idx_name)

    def _find(self, cls, query):
        coll = self.get_collection(cls.get_table_name())

        final_results = []
        for db_result in coll.find(query):
            # We need to give JSON to from_data
            obj_data = json.dumps(db_result['value'])
            obj = cls.from_data(obj_data)
            final_results.append(obj)

        return final_results

    def find_one(self, cls, id):
        one = self._find(cls, {"_id": id})
        if not one:
            return None
        return one[0]

    def find_all(self, cls):
        return self._find(cls, {})

    def find_by_index(self, cls, index_name, value):
        return self._find(cls, {index_name: str(value)})

    def save(self, obj):
        if not obj.id:
            obj.id = uuid()

        stored_data = {
            '_id': obj.id,
            'value': json.loads(obj.to_data())
        }

        index_vals = obj.indexes() or {}
        for key in obj.__class__.index_names() or []:
            val = index_vals.get(key, '')
            stored_data[key] = str(val)

        coll = self.get_collection(obj.__class__.get_table_name())
        coll.update({"_id": obj.id}, stored_data, upsert=True)
